open Graphics
open Types

let debug = false

type coord = int (* 0<= n < 600 *)
type pos = coord*coord
type component = int (* 0<= n < 256 *)
type rgb = component*component*component
type transparency = component
type pixel = rgb * transparency
type bitmap = pixel array array (* 600x600 *)
type color =  Rgb of rgb
	     | Transparency of transparency
type bucket =  color list
type dir = N | E | S | W

let black = 0,0,0
let red = 255,0,0
let green = 0,255,0
let yellow = 255,255,0
let blue = 0,0,255
let magenta = 255,0,255
let cyan = 0,255,255
let white =  255,255,255
let transparent = 0
let opaque = 255

let rblack = Rgb black
let rred = Rgb red
let rgreen = Rgb green
let ryellow = Rgb yellow
let rblue = Rgb blue
let rmagenta = Rgb magenta
let rcyan = Rgb cyan
let rwhite = Rgb white
let rtransparent = Transparency transparent
let ropaque = Transparency opaque

let transparentBitmap () = 
  Array.make_matrix 600 600 (black, transparent)
(* State *)
let (bucket:bucket ref) = ref [] 
let (position:pos ref)  = ref (0,0)
let (mark:pos ref)  = ref (0,0)
let (dir:dir ref) = ref E
let (bitmaps: bitmap list ref) = 
    ref [transparentBitmap ()]

let print_dir fmt = function
  | E -> Format.fprintf fmt "E"
  | S -> Format.fprintf fmt "S"
  | W -> Format.fprintf fmt "W"
  | N -> Format.fprintf fmt "N"

let dump_state () = 
  Format.printf "Buckets: %d@\nPosition:%d,%d@\nmarck:%d,%d@\ndir:%a@\nbitmaps:%d@." (List.length !bucket) (fst !position) (snd !position) (fst !mark) (snd !mark) print_dir !dir (List.length !bitmaps)

let reset () = 
  bucket := [];
  position := (0,0);
  mark := (0,0);
  dir := E;
  bitmaps := [transparentBitmap ()]

let () = open_graph " 600x600"
let () = set_window_title "Brought to you by The Caml Riders"

let draw ?(verbose=0) ?(offset=0) bitmap  = 
  let ml_color_array = Array.init 600 (fun x -> Array.make 600 Graphics.black) in
  for x=0 to 599 do
    for y=0 to 599 do
      ml_color_array.(y).(x) <- let ((r,g,b),_) = bitmap.(x).(y) in rgb r g b
    done
  done;
  let image = make_image ml_color_array in
  draw_image image offset 0

let average l default = 
  match l with 
    | [] -> default
    | l -> 
	let count = ref 0 in
	let sum = ref 0 in
	List.iter
	  (fun c -> sum := c + !sum; incr count) 
	  l;
	!sum / !count

let t = ref (Unix.time());;

let draw_top ?(force=true) ?(verbose=0) () = 
  let t' = Unix.time () in
  if not force && t' -. !t < 0.1 then () else
    begin
      t := t';
      match !bitmaps with
	| [b] -> draw b
	| b1::b2::_ -> draw ~verbose b1 ; 
	    (* draw ~verbose ~offset:600 b2 *)
	| [] -> ()
    end
;;

let rec split_colors ((r,g,b,a) as acc) bucket = 
  match bucket with 
    | [] -> acc
    | Rgb (r0,g0,b0)::remain -> split_colors (r0::r,g0::g,b0::b,a) remain
    | Transparency a0::remain -> split_colors (r,g,b,a0::a) remain


let curPixel = ref ((0,0,0),0)
let curPixelOK = ref false
  
let currentPixel () = 
  if !curPixelOK then !curPixel else
  let r, g, b ,a = split_colors ([],[],[],[]) !bucket in
  let rc = average r 0 in
  let gc = average g 0 in
  let bc = average b 0 in
  let ac = average a 255 in
  curPixel := ((rc*ac)/255,(gc*ac)/255,(bc*ac)/255) ,ac;
  curPixelOK := true;
  !curPixel

let getPixel (x,y) = 
  (List.hd !bitmaps).(x).(y)

let setPixel (x,y) = 
  (*if debug then Format.printf "Setting %d,%d@." x y;*)
  (List.hd !bitmaps).(x).(y) <- currentPixel ()

let addColor c = 
  curPixelOK := false;
  bucket := c::!bucket

let move (x,y) dir = 
  (*if debug then Format.printf "Moving from %d,%d@." x y;*)
  match dir with
  | N -> x, (y+599) mod 600
  | E -> (x+1) mod 600, y
  | S -> x, (y+1) mod 600
  | W -> (x+599) mod 600, y

let turnCounterClockwise dir = 
  (*if debug then Format.printf "Turning CCW@.";*)
  match dir with
  | N -> W
  | E -> N
  | S -> E
  | W -> S

let turnClockwise dir = 
  (*if debug then Format.printf "Turning CW@.";*)
  match dir with
  | N -> E
  | E -> S
  | S -> W
  | W -> N

let line (x0,y0) (x1,y1) = 
  (*if debug then Format.printf "Line %d,%d -> %d,%d@." x0 y0 x1 y1;*)
  let deltax = x1-x0 in
  let deltay = y1-y0 in
  let d = max (abs deltax) (abs deltay) in
  let c = if deltax*deltay <= 0 then 1 else 0 in
  let x = ref (x0*d + (d-c)/2) in 
  let y = ref (y0*d + (d-c)/2) in
  for i = 1 to d do
    setPixel (!x/d,!y/d);
    x:=!x+deltax;
    y:=!y+deltay;
  done;
  setPixel (x1,y1)
      
let rec fill l initial = 
  match l with
    | [] -> ()
    | (x,y as pos)::acc ->
	if getPixel pos = initial then
	  begin 
	    setPixel pos;
	    let acc = if y<599 then (x,y+1)::acc else acc in
	    let acc = if y>0 then (x,y-1)::acc else acc in
	    let acc = if x<599 then (x+1,y)::acc else acc in
	    let acc = if x>0 then (x-1,y)::acc else acc in
	    fill acc initial
	  end
	else
	  fill acc initial
  
let tryfill () = 
  (*if debug then Format.printf "Tryfill@."; *)
  let new_ = currentPixel () in
  let old = getPixel !position in
  if new_ <> old then fill [!position] old

let addBitmap b = 
  (*if debug then Format.printf "addBitmap@."; *)
  if List.length !bitmaps < 10 then 
    bitmaps := b::!bitmaps

let compose () = 
  (*if debug then Format.printf "Compose@."; *)
  match !bitmaps with
    | bitmaps0::((bitmaps1::remain) as next) -> 
	for x=0 to 599 do
	  for y = 0 to 599 do
	    let (r0,g0,b0),a0 = bitmaps0.(x).(y) in
	    let (r1,g1,b1),a1 = bitmaps1.(x).(y) in

	    let calc arg = (arg*(255-a0))/255 in

	    bitmaps1.(x).(y) <- ((r0+calc r1,
				  g0+calc g1,
				  b0+calc b1),
				 a0+calc a1)
	  done
	done;
	bitmaps := next
    | _ -> ()

let clip () = 
  (*if debug then Format.printf "Clip@."; *)
   match !bitmaps with
    | bitmaps0::((bitmaps1::remain) as next) -> 
	for x=0 to 599 do
	  for y = 0 to 599 do
	    let (r0,g0,b0),a0 = bitmaps0.(x).(y) in
	    let (r1,g1,b1),a1 = bitmaps1.(x).(y) in

	    bitmaps1.(x).(y) <- ((r1*a0/255,
				  g1*a0/255,
				  b1*a0/255),
				 a1*a0/255)
	  done
	done;
	bitmaps := next
    | _ -> ()

(*
let exec0 r = match r with
  | [| P;I;P;I;I;I;C |] -> addColor rblack
  | [| P;I;P;I;I;I;P |] -> addColor rred
  | [| P;I;P;I;I;C;C |] -> addColor rgreen
  | [| P;I;P;I;I;C;F |] -> addColor ryellow
  | [| P;I;P;I;I;C;P |] -> addColor rblue
  | [| P;I;P;I;I;F;C |] -> addColor rmagenta
  | [| P;I;P;I;I;F;F |] -> addColor rcyan
  | [| P;I;P;I;I;P;C |] -> addColor rwhite
  | [| P;I;P;I;I;P;F |] -> addColor rtransparent
  | [| P;I;P;I;I;P;P |] -> addColor ropaque

  | [| P;I;I;P;I;C;P |] -> bucket := []
  | [| P;I;I;I;I;I;P |] -> position := move !position !dir
  | [| P;C;C;C;C;C;P |] -> dir := turnCounterClockwise !dir
  | [| P;F;F;F;F;F;P |] -> dir := turnClockwise !dir
  | [| P;C;C;I;F;F;P |] -> mark := !position
  | [| P;F;F;I;C;C;P |] -> line !position !mark; draw (List.hd !bitmaps)
  | [| P;I;I;P;I;I;P |] -> tryfill () ; draw (List.hd !bitmaps)
  | [| P;C;C;P;F;F;P |] -> addBitmap (transparentBitmap ()); draw (List.hd !bitmaps)
  | [| P;F;F;P;C;C;P |] -> compose (); draw (List.hd !bitmaps)
  | [| P;F;F;I;C;C;F |] -> clip (); draw (List.hd !bitmaps)
  | _ -> ()
      *)

let db ~verbose c = if verbose >= 1 then Format.printf "%c@?" c
exception BUILD
let exec ?(verbose=0) ~drawing r = 
(*  let verbose = 2 in*)
  match r with
| [| P;I;P;I;I;I;C |] -> addColor rblack;
    db ~verbose 'C';true
| [| P;I;P;I;I;I;P |] -> addColor rred;
    db ~verbose 'C';true
| [| P;I;P;I;I;C;C |] -> addColor rgreen;
    db ~verbose 'C';true
| [| P;I;P;I;I;C;F |] -> addColor ryellow;
    db ~verbose 'C';true
| [| P;I;P;I;I;C;P |] -> addColor rblue;
    db ~verbose 'C';true
| [| P;I;P;I;I;F;C |] -> addColor rmagenta;
    db ~verbose 'C';true
| [| P;I;P;I;I;F;F |] -> addColor rcyan;
    db ~verbose 'C';true
| [| P;I;P;I;I;P;C |] -> addColor rwhite;
    db ~verbose 'C';true
| [| P;I;P;I;I;P;F |] -> addColor rtransparent;
    db ~verbose 'T';true
| [| P;I;P;I;I;P;P |] -> addColor ropaque;
    db ~verbose 'O';true
| [| P;I;I;P;I;C;P |] -> bucket := [];
    db ~verbose 'E';true
| [| P;I;I;I;I;I;P |] -> position := move !position !dir;
    db ~verbose 'M';true
| [| P;C;C;C;C;C;P |] -> dir := turnCounterClockwise !dir;
    db ~verbose 'T';true
| [| P;F;F;F;F;F;P |] -> dir := turnClockwise !dir;
    db ~verbose 'W';true
| [| P;C;C;I;F;F;P |] -> mark := !position;
    db ~verbose '*';true
| [| P;F;F;I;C;C;P |] -> 
    line !position !mark;
    db ~verbose 'L';if drawing then draw_top ();true
| [| P;I;I;P;I;I;P |] -> tryfill ();
    db ~verbose 'F';
    if drawing then draw_top ();true
| [| P;C;C;P;F;F;P |] -> 
    addBitmap (transparentBitmap ());
    db ~verbose 'A';
    if drawing then draw_top ();true
| [| P;F;F;P;C;C;P |] -> compose ();
    db ~verbose '$';
    if drawing then draw_top ();true
| [| P;F;F;I;C;C;F |] -> clip ();
    db ~verbose '%';
    if drawing then draw_top ();true
(*
| [| P;P;P;P;P;P;P |] ->
    (* if verbose = 666 then raise Misc.Dumpstack; *) false
*)
| [| _;_;_;_;_;_;_ |] -> 
    if verbose >= 2 then Format.printf "@.Skipping non drawable rna7: %a@." Misc.print_rna7 r;
    if debug then raise BUILD;
    false
| _ -> assert false

let build ?(verbose=0) rna = 
  List.iter (fun x -> ignore (exec ~verbose:verbose ~drawing:false x)) rna;
  draw_top ()

(*
(* test *)
let b = rblack
let r = rred
let m = rmagenta
let w = rwhite
let y = ryellow
let c = rcyan
let t = rtransparent
let o = ropaque

let () = assert (reset ();
		 bucket := [t ; o; o ];
		 currentPixel()=((0,0,0),170));
  assert (reset ();
	  bucket := [b;y;c;];
	  currentPixel()=((85,170,85),255));
  assert (reset ();
	  bucket := [y;t;o;];
	  currentPixel()=((127,127,0),127));
  assert (reset ();
	  bucket := [b;b;b;b;b;b;b;b;b;b;b;b;b;b;b;b;b;b;r;r;r;r;r;r;r;
		     m;m;m;m;m;m;m;m;m;m;m;m;m;
		     m;m;m;m;m;m;m;m;m;m;m;m;m;
		     m;m;m;m;m;m;m;m;m;m;m;m;m;
		     w;w;w;w;w;w;w;w;w;w;
		     o;o;o;
		     t;];
	  currentPixel()=((143,25,125),191));
  assert (reset ();
	  let x,y = move (100,0) N in
	  (*Format.printf "x:%d y:%d@." x y;*)
	   (x,y) = (100,599)
	 );
  reset()  
*)
