
open Format
open Types
open Misc

type p_item =
  | Base of base | Bang of int | Ques of base list | Left | Right

type pattern = p_item list (* not reversed *)

let print_base fmt b =
  fprintf fmt "%c" (Misc.char_of_base b)

let print_p_item fmt = function
  | Base b -> fprintf fmt "%a" Misc.print_base b
  | Bang n -> fprintf fmt "!%d" n
  | Ques s -> fprintf fmt "?<"; List.iter (print_base fmt) s; fprintf fmt ">"
  | Left -> fprintf fmt "("
  | Right -> fprintf fmt ")"

let print_pattern fmt pl = List.iter (print_p_item fmt) pl

let stack_dump dna =
  Format.eprintf "\n-----------STACK DUMP-------------\n";
  let n = (Dna.search_forward dna 0 [I;F;P;I;C;F;P;P;C;F;I;P;P]) in
  Format.eprintf "Size:%d(from %d)\n" ((Dna.len dna)-n) n;
  (try
     for i=n to n+40 do
       let b = Dna.get dna i in
	Format.eprintf "%a" Misc.print_base b;
      done;
       Format.eprintf "...";
    with Not_found -> ());
    Format.eprintf "\n-----------END STACK DUMP-------------\n";
    Format.eprintf "@."


exception Finish

let finish dna = (*stack_dump dna;*) raise Finish

let overflow n =
  if n < 0 then Pervasives.max_int else n

let rec nat dna = 
  try
    let next,dna = Dna.pop dna in
    match next with
      | P -> dna,0
      | I | F -> let dna,n = nat dna in dna, overflow(n+n)
      | C -> let dna,n = nat dna in dna,overflow(n+n+1)
  with Not_found -> finish dna

let rec consts dna =
  try
    let next,dna' = Dna.pop dna in
    match next with
      | C -> let dna',s = consts dna' in dna',(I :: s) 
      | F -> let dna',s = consts dna' in dna',(C :: s) 
      | P -> let dna',s = consts dna' in dna',(F :: s) 
      | I  -> 
	  let next,dna' = Dna.pop dna' in
	  match next with
	    | C -> let dna',s = consts dna' in dna',(P :: s) 
	    | _ -> dna,[]
  with Not_found -> dna,[]


(* DONE fixme: page 12 le dernier RNA peut etre plus court  *)
let apply_rna dna k =
  let rna = Array.create 7 I in
  let rec get_rna dna i =
    if i = 7 then dna else
      let next,dna = try Dna.pop dna with Not_found -> I,dna in
      rna.(i) <- next;
      get_rna dna (i+1)
  in
  let dna = get_rna dna 0 in
  (try
    k rna
  with Misc.Dumpstack -> 
    (* display stack *)
    stack_dump dna
  )
  ; dna

let pattern dna k =
  let rec step dna p lvl =
    let next,dna = Dna.pop dna in
    match next with
      | C -> step dna (Base I :: p) lvl
      | F -> step dna (Base C :: p) lvl
      | P -> step dna (Base F :: p) lvl
      | I ->
	  let next,dna = Dna.pop dna in
	  match next with
	    | C -> step dna (Base P :: p) lvl
	    | P -> 
		let dna,n = nat dna in
		step dna (Bang n :: p) lvl
	    | F -> 
		let _,dna = Dna.pop dna in (* three bases consumed see page 6 *)
		let dna,s = consts dna in
		step dna (Ques s :: p) lvl
	    | I ->
		let next,dna = Dna.pop dna in
		match next with
		  | P -> step dna (Left :: p) (lvl+1)
		  | C | F ->
		      if lvl = 0 then
			dna,List.rev p
		      else
			step dna (Right :: p) (lvl-1)
		  | I ->
		      let dna = apply_rna dna k in
		      step dna p lvl
  in 
  try 
    step dna [] 0    
  with Not_found -> finish dna


type t_item = 
  | TBase of base | Index of int * int | Bar of int
type template = t_item list

let print_t_item fmt i =
  match i with
    | TBase b -> print_base fmt b
    | Index(n,l) -> fprintf fmt "%d_{%d}" n l
    | Bar(n) -> fprintf fmt "|%d|" n

let print_template fmt pl = List.iter (print_t_item fmt) pl

let template ?(verbose=0) n dna k =
  let rec step m dna t =
    if verbose=2 then eprintf "Template%d_%d@." n m;
    let next,dna = Dna.pop dna in
    match next with
      | C -> if verbose=2 then eprintf "C@."; step (m+1) dna (TBase I :: t) 
      | F -> if verbose=2 then eprintf "F@."; step (m+1) dna (TBase C :: t) 
      | P -> if verbose=2 then eprintf "P@."; step (m+1) dna (TBase F :: t) 
      | I ->
	  let next,dna = Dna.pop dna in
	  match next with
	    | C -> if verbose=2 then eprintf "IC@."; step (m+1) dna (TBase P :: t)
	    | P | F -> 
		let dna,l = nat dna in
		let dna,n = nat dna in
		if verbose=2 then eprintf "IP/IF@."; step (m+1) dna (Index(n,l) :: t)
	    | I ->
		let next,dna = Dna.pop dna in
		match next with
		  | C | F -> if verbose=2 then eprintf "IIC@."; dna, List.rev t
		  | P -> 
		      if verbose=2 then eprintf "IIP@."; let dna,n = nat dna in
		      step (m+1) dna (Bar n :: t)
		  | I ->
		      if verbose=2 then eprintf "III@."; let dna = apply_rna dna k in
		      step (m+1) dna t
  in    
  try
    step 0 dna [] 
  with Not_found -> finish dna

(* page 11 *)

let rec quote d =
  match d with
    | [] -> []
    | h::t ->
	match h with
	  | I -> C :: quote t
	  | C -> F :: quote t
	  | F -> P :: quote t
	  | P -> I :: C :: quote t

let rec protect_list l d =
  if l=0 then d else protect_list (l-1) (quote d)

let protect l d =
  if l=0 then d else Dna.of_list (protect_list l (Dna.to_list d))

let rec asnat n =
  if n=0 then [P] else
    if n mod 2 = 0 then I :: (asnat (n/2)) else
      C :: (asnat (n/2))

type environment = Dna.t array

(* page 10 *)
let replace dna (tpl:template) (e:environment) =
  let r =
    List.fold_left
      (fun r t ->
	 match t with
	   | TBase b -> Dna.append r (Dna.base b)
	   | Index(n,l) -> 
	       begin
		 let en = 
		   try
		     e.(n)
		   with
		       Invalid_argument _ -> Dna.empty
		 in
		 Dna.append r (protect l en)
	       end
	   | Bar n ->
	       begin
		 let ln = 
		   try
		     Dna.len (e.(n))
		   with
		       Invalid_argument _ -> 0
		 in
		 Dna.append r (Dna.of_list (asnat ln))
	       end)
      Dna.empty 
      tpl
  in
  Dna.append r dna

(* page 9 *)
let matchreplace dna pat t =
  let step i e c p =
    match p with
      | Base b ->
	  (try
	    if Dna.get dna i = b then (i+1,e,c) else raise Exit
	  with Not_found -> assert false)
      | Bang n ->
	  let i = i + n in
	  if n = Pervasives.max_int || i < 0 || i > Dna.len dna then raise Exit
	  else (i,e,c)
      | Ques s ->
	  begin
	    try
	      (*eprintf "Before search: i=%d@." i; *)
	      let n = i + (Dna.search_forward dna i s) in
	      (*eprintf "Success: n=%d@." n;*)
	      (n,e,c)
	    with Not_found -> 
	      (* eprintf "Search failed@."; *)
	      raise Exit
	  end
      | Left -> (i,e,i::c)	  
      | Right -> 
	  begin match c with
	    | [] -> assert false
	    | c0 :: c1 ->
		(i,(Dna.sub dna c0 i)::e,c1)
	  end
  in
  try
    let (i,e,c) =
      List.fold_left
	(fun (i,e,c) p -> step i e c p)
	(0,[],[])
	pat
    in
    let dna' = Dna.skip dna i in
    let l = List.length e in
    let ea = Array.create l Dna.empty in
    let _ = List.fold_left
      (fun i x -> ea.(i) <- x; i-1)
      (l-1) e 
    in
    replace dna' t ea
  with Exit -> dna

let one_step ?(verbose=0) n dna k =
  if verbose=2 then
    Format.eprintf "Before Pattern:'%a'@." (Dna.print_start 20) dna;
  let dna,p = pattern dna k in
  if verbose=1 then eprintf "pattern%d: @[%a@]@." n (Pp.print_list Pp.nothing print_p_item) p;
  if verbose=2 then
    Format.eprintf "Before Template:'%a'@." (Dna.print_start 2000) dna;
  let dna,t = template ~verbose n dna k in
  if verbose=1 then 
    eprintf "template%d: @[%a@]@." n (Pp.print_list Pp.nothing print_t_item) t;
  let r = matchreplace dna p t in
  if verbose=2 then
    begin
      Format.eprintf "Before MatchReplace:'%a'@." (Dna.print_start 2000) dna;
      Format.eprintf "After MatchReplace:'%a'@." (Dna.print_start 2000) r;
    end;
  r

let execute ?(verbose=0) dna k =
  let rec rec_exec n dna k =
    let dna = one_step ~verbose n dna k in
    (*stack_dump dna;*)
    rec_exec (n+1) dna k
  in 
  try
    rec_exec 0 dna k
  with
      Finish ->     stack_dump dna; ()


(*

let test n t r =
  let dna = Dna.of_string t in
  let dna = one_step dna
    (fun rna ->
       for i=0 to 6 do
	 printf "RNA: ";
	 printf "%a" Misc.print_base rna.(i);
	 printf "@.";
       done)
  in 
  let dna = Dna.sub dna ~start:0 ~stop:100000 in
  if dna = Misc.baselist_of_string r then 
    printf "test %d ok@." n
  else 
    begin
      printf "FAILED! got ";
      List.iter
	(fun b -> printf "%a" Misc.print_base b) dna;
      printf " instead of %s@." r;
    end

let test150 = "IIPIPICPIICICIIFICCIFPPIICCFPC"
let res150 = "PICFC"

let () = test 150 test150 res150

let test151="IIPIPICPIICICIIFICCIFCCCPPIICCFPC"
let res151 = "PIICCFCFFPC"

let () = test 151 test151 res151

let test152 = "IIPIPIICPIICIICCIICFCFC"
let res152 ="I"

let () = test 152 test152 res152
*)


(*

  Build.reset ();
  let dna = Dna.create ~prefix:Sys.argv.(1) ~dna:Sys.argv.(2) in
  let t = ref (Unix.time()) in
  execute dna 
    (fun rna -> 
       let drawing = true || 
         let t' = Unix.time() in
         if t' > !t +. 1.0 then (t:=t';true) else false in
       Build.exec ~drawing rna;);


*)
