
open Format
open Types
open Misc
open Unix

(****

type t = (string * int * int) list    
    
let print fmt t = 
  List.iter (fun (s,i,l) -> Format.fprintf fmt "%s" (String.sub s i l)) t

let empty = []

let of_string s = [s,0,String.length s]

let base b = of_string (String.make 1 (char_of_base b))

let rec length_rec t len =
  match t with
      [] -> len
    | (s,pos,slen) :: tail -> 
        length_rec tail (len+slen)
	  
let len t = length_rec t 0

let create prefix dna = 
  let lprefix = (stat prefix).st_size in
  let ldna = (stat dna).st_size in
  let n = lprefix + ldna in
  assert (n < Sys.max_string_length);
  let s = String.create n in
  let rec fill i cin = 
    match (try Some (input_char cin) with End_of_file -> None) with
      | Some c -> s.[i] <- c; fill (i+1) cin
      | None -> i
  in
  let m = fill (fill 0 (open_in prefix)) (open_in dna) in
  assert (m = n);
  of_string s

let to_string s =
  let b = Buffer.create 1000 in
  List.iter (fun (s,pos,len) ->
               Buffer.add_substring b s pos len
	    ) s;
  Buffer.contents b
    
let rec get list pos =
  match list with
      [] -> raise Not_found
    | (s,spos,slen) :: tail ->
        if pos >= slen then
          get tail (pos - slen)
        else
          base_of_char s.[spos + pos]
	    
let rec skip s n =
  match s with
      [] -> []
    | (s,pos,len) :: tail ->
        if len <= n then
          skip tail (n-len)
        else
          (s,pos+n,len-n) :: tail

let pop s = get s 0, skip s 1
	    
let rec sub_rec s len list =
  match s with
      [] -> List.rev list
    | (s,pos,slen) :: tail ->
        if slen <= len then
          sub_rec tail (len-slen) ( (s,pos,slen) :: list )
        else
          List.rev ((s,pos,len) :: list)
    
let sub s start stop =
  (*printf "sub start=%d stop=%d@." start stop;*)
  if stop <= start then 
    empty
  else   
    let s = skip s start in
    sub_rec s (stop-start) []

(***
let sub t ~start ~stop =
  printf "sub start=%d stop=%d =>" start stop;
  let s = sub t ~start ~stop in
  if stop-start <= 50 then printf "%a@." print s else printf "<...>@.";
  s
***)
    
let subs s pos len =
  if pos = 0 && len = String.length s then s else
    String.sub s pos len
      
let append s d =
  let rec iter s d =
    match s with
        [] -> d
      | ((s1,pos1,len1) as c) :: tail -> 
          match d with
              [] -> assert false
            | (s2, pos2,len2) :: tail2 ->
                if len1 + len2 < 10000 then
                  iter tail (
                    (subs s1 pos1 len1 ^ subs s2 pos2 len2,
                     0, len1+len2) :: tail2)
                else
                  iter tail (c :: d)
  in
  if d=[] then s else iter (List.rev s) d

(*
let append x y =
  let r = append x y in
  printf "append %d + %d = %d@." (len x) (len y) (len r);
  r
*)

module T = struct
  type t = (string * int * int) list
  let length = len
  type char = base
  let get t i = get t i
end
module P = struct
  type t = base array
  let length = Array.length
  type char = base
  let get = Array.get
end
module KnuthMorrisPratt = Kmp.Make(T)(P)

let search_forward t i s = 
  let t = skip t i in
  let p = Array.of_list s in
  let index = KnuthMorrisPratt.search t p in
  index + Array.length p

let to_list t =
  let rec scan acc s n i =
    if i >= n then acc else scan (base_of_char s.[i] :: acc) s n (i+1)
  in
  List.rev (List.fold_left (fun acc (s,i,l) -> scan acc s (i+l) i) [] t)

let of_list bl =
  let s = String.create (List.length bl) in
  let rec fill i = function
    | [] -> of_string s
    | b :: bl -> s.[i] <- char_of_base b; fill (i+1) bl
  in
  fill 0 bl

let print_start length fmt dna = 
  let t = sub dna 0 length in
  let l = to_list t in
  List.iter (fun  b -> print_base fmt b) l 

*****)

(* ropes *)

type t = 
    | Str of string * int * int (* 0 <= ofs < len(s), ofs+len <= len(s) *)
    | App of t * t * int        (* 0 < len t1, len t2 *)

let empty = Str ("", 0, 0)

let len = function
  | Str (_,_,n) 
  | App (_,_,n) -> n

let of_string s = Str (s, 0, String.length s)

let create prefix dna = 
  let lprefix = (stat prefix).st_size in
  let ldna = (stat dna).st_size in
  let n = lprefix + ldna in
  assert (n < Sys.max_string_length);
  let s = String.create n in
  let rec fill i cin = 
    match (try Some (input_char cin) with End_of_file -> None) with
      | Some c -> s.[i] <- c; fill (i+1) cin
      | None -> i
  in
  let m = fill (fill 0 (open_in prefix)) (open_in dna) in
  assert (m = n);
  of_string s

let small_length = 10000
let is_small s = String.length s < small_length

(* smart constructor *)
let app = function
  | Str (_,_,0), t | t, Str (_,_,0) -> 
      t
  | Str (s1, ofs1, len1), Str (s2, ofs2, len2) 
    when len1 <= small_length && len2 <= small_length ->
      Str (String.sub s1 ofs1 len1 ^ String.sub s2 ofs2 len2, 0, len1 + len2)
  | App (t1, Str (s1, ofs1, len1), _), Str (s2, ofs2, len2)
    when len1 <= small_length && len2 <= small_length ->
      App (t1, 
	   Str 
	     (String.sub s1 ofs1 len1 ^ String.sub s2 ofs2 len2, 0, len1+len2),
	   len t1 + len1 + len2)
  | Str (s1, ofs1, len1), App (Str (s2, ofs2, len2), t2, _)
    when len1 <= small_length && len2 <= small_length ->
      App (Str 
	     (String.sub s1 ofs1 len1 ^ String.sub s2 ofs2 len2, 0, len1+len2),
	   t2, len1 + len2 + len t2)
  | t1, t2 -> 
      App (t1, t2, len t1 + len t2)

let rec get t i = match t with
  | Str (s, ofs, len) ->
      if i >= len then raise Not_found;
      base_of_char s.[ofs + i]
  | App (t1, t2, _) ->
      let n1 = len t1 in
      if i < n1 then get t1 i else get t2 (i-n1)

let rec pop = function
  | Str (_, _, 0) -> 
      raise Not_found
  | Str (s, ofs, len) ->
      base_of_char s.[ofs], if len = 1 then empty else Str (s, ofs+1, len-1)
  | App (t1, t2, _) ->
      let b, t1 = pop t1 in
      b, app (t1, t2)

let rec skip t i = match t with
  | Str (s, ofs, len) ->
      if i >= len then empty else Str (s, ofs+i, len-i)
  | App (t1, t2, _) ->
      let n1 = len t1 in
      if i >= n1 then skip t2 (i-n1) else app (skip t1 i, t2)

(* assumption : start < stop <= len(t) *)
let rec mksub start stop = function
  | Str (s, ofs, len) -> 
      Str (s, ofs+start, stop-start)
  | App (t1, t2, _) ->
      let n1 = len t1 in
      if stop <= n1 then mksub start stop t1 
      else if start >= n1 then mksub (start-n1) (stop-n1) t2
      else app (mksub start n1 t1, mksub 0 (stop-n1) t2)

let sub t start stop = 
  let stop = min stop (len t) in
  if stop <= start then empty else mksub start stop t

module T = struct
  type dna = t 
  type t = dna (* grrr *)
  let length = len
  type char = base
  let get = get
end
module P = struct
  type t = base array
  let length = Array.length
  type char = base
  let get = Array.get
end
module KnuthMorrisPratt = Kmp.Make(T)(P)

let search_forward t i s = 
  let t = skip t i in
  let p = Array.of_list s in
  let index = KnuthMorrisPratt.search t p in
  index + Array.length p

let append t1 t2 = app (t1, t2)

let of_list bl =
  let s = String.create (List.length bl) in
  let rec fill i = function
    | [] -> of_string s
    | b :: bl -> s.[i] <- char_of_base b; fill (i+1) bl
  in
  fill 0 bl

let base b = of_list [b]

let to_list t =
  let rec scan acc i =
    if i < 0 then acc else scan (get t i :: acc) (i-1)
  in
  scan [] (len t - 1)

let print_start length fmt dna = 
  let t = sub dna 0 length in
  let l = to_list t in
  List.iter (fun  b -> print_base fmt b) l 

let print fmt t =
  fprintf fmt "..."

(* log rope sizes *)

(****

let max_size = ref min_int
let tot_size = ref 0.0
let nb_sizes = ref 0
let rec size = function
  | Str _ -> 0
  | App (r1, r2, _) -> 1 + size r1 + size r2
let log_size r =
  let s = size r in
  if s > !max_size then max_size := s;
  tot_size := !tot_size +. float s;
  incr nb_sizes;
  r
let () = at_exit 
  (fun () -> 
    eprintf "max size = %d@." !max_size;
    eprintf "avg size = %f@." (!tot_size /. float !nb_sizes);
    eprintf "nb logs  = %d@." !nb_sizes)

let skip r n = log_size (skip r n)
let append r1 r2 = log_size (append r1 r2)
let sub t i j = log_size (sub t i j)

****)

(* Functorial code *)

module type STRING = sig
  type t

  val empty : t
    
  val add : char -> t -> t
    
  val len : t -> int

  val sub : t -> int -> int -> t

  val app : t -> t -> t

  val get : t -> int -> char
end

module type ROPE = sig
  type t

  val empty : t
    
  val add : string -> t -> t
    
  val len : t -> int

  val sub : t -> int -> int -> t

  val app : t -> t -> t

  val get : t -> int -> char
end

module RopeOfString(S : STRING) : ROPE = struct

  include S

  let of_string s =
    let rec mk i t = if i < 0 then t else mk (i-1) (S.add s.[i] t) in
    mk (String.length s - 1) S.empty

  let add s = S.app (of_string s)

end

module Make(S : ROPE) = struct

  type t = S.t

  let empty = S.empty

  let of_string s = S.add s S.empty

  let create prefix dna = 
    let lprefix = (stat prefix).st_size in
    let ldna = (stat dna).st_size in
    let n = lprefix + ldna in
    assert (n < Sys.max_string_length);
    let s = String.create n in
    let rec fill i cin = 
      match (try Some (input_char cin) with End_of_file -> None) with
	| Some c -> s.[i] <- c; fill (i+1) cin
	| None -> i
    in
    let m = fill (fill 0 (open_in prefix)) (open_in dna) in
    assert (m = n);
    of_string s

  let len = S.len

  let get s i = base_of_char (S.get s i)

  let skip s i =
    if i >= S.len s then empty else S.sub s i (S.len s)

  let pop s =
    if len s = 0 then raise Not_found;
    base_of_char (S.get s 0), skip s 1

  let append = S.app

  let sub t start stop = 
    let stop = min stop (len t) in
    if stop <= start then empty else S.sub t start stop

  module T = struct
    type dna = t 
    type t = dna (* grrr *)
    let length = len
    type char = base
    let get = get
  end
  module P = struct
    type t = base array
    let length = Array.length
    type char = base
    let get = Array.get
  end
  module KnuthMorrisPratt = Kmp.Make(T)(P)
    
  let search_forward t i s = 
    let t = skip t i in
    let p = Array.of_list s in
    let index = KnuthMorrisPratt.search t p in
    index + Array.length p
      
  let of_list bl =
    let s = String.create (List.length bl) in
    let rec fill i = function
      | [] -> of_string s
      | b :: bl -> s.[i] <- char_of_base b; fill (i+1) bl
    in
    fill 0 bl

  let base b = of_list [b]

  let to_list t =
    let rec scan acc i =
      if i < 0 then acc else scan (get t i :: acc) (i-1)
    in
    scan [] (len t - 1)

  let print_start length fmt dna = 
    let t = sub dna 0 length in
    let l = to_list t in
    List.iter (fun  b -> print_base fmt b) l 

  let print fmt t =
    fprintf fmt "..."

end

module S : STRING = struct

  type t = string

  let empty = ""
    
  let add c s = (String.make 1 c) ^ s
    
  let len = String.length

  let sub s start stop = String.sub s start (stop - start)

  let app = (^)

  let get s i = String.unsafe_get s i

end

module R : ROPE = struct

  type t = (string * int * int) list    
    
  let empty = []

  let of_string s = [s,0,String.length s]

  let rec length_rec t len =
    match t with
	[] -> len
      | (s,pos,slen) :: tail -> 
          length_rec tail (len+slen)
	    
  let len t = length_rec t 0

  let rec get list pos =
    match list with
	[] -> raise Not_found
      | (s,spos,slen) :: tail ->
          if pos >= slen then
            get tail (pos - slen)
          else
            s.[spos + pos]
	    
  let rec sub_rec s len list =
    match s with
	[] -> List.rev list
      | (s,pos,slen) :: tail ->
          if slen <= len then
            sub_rec tail (len-slen) ( (s,pos,slen) :: list )
          else
            List.rev ((s,pos,len) :: list)
	      
  let rec skip s n =
    match s with
	[] -> []
      | (s,pos,len) :: tail ->
          if len <= n then
            skip tail (n-len)
          else
            (s,pos+n,len-n) :: tail

  let sub s start stop =
    (*printf "sub start=%d stop=%d@." start stop;*)
    if stop <= start then 
      empty
    else   
      let s = skip s start in
      sub_rec s (stop-start) []
	
  let subs s pos len =
    if pos = 0 && len = String.length s then s else
      String.sub s pos len
	
  let append s d =
    let rec iter s d =
      match s with
          [] -> d
	| ((s1,pos1,len1) as c) :: tail -> 
            match d with
		[] -> assert false
              | (s2, pos2,len2) :: tail2 ->
                  if len1 + len2 < 10000 then
                    iter tail (
                      (subs s1 pos1 len1 ^ subs s2 pos2 len2,
                      0, len1+len2) :: tail2)
                  else
                    iter tail (c :: d)
    in
    if d=[] then s else iter (List.rev s) d
      
  let app = append

  let add s t = append (of_string s) t

end

(*include Make(R)*)
(*include Make(Strings.Rope) (* Assert_failure("strings.ml", 88, 2) *)*)
(*include Make(RopeOfString(Strings.MyString)) (* exception Not_found *)*)

(*include Make(Rope.S)*)








