
open Types
open Misc
open Unix

type t = 
    | String of int * string
    | Append of int * base list * t

let empty = String (0, "")

let rec len = function
  | String (i, s) -> String.length s - i 
  | Append (n, _, t) -> n + len t

let create ~prefix ~dna = 
  let lprefix = (stat prefix).st_size in
  let ldna = (stat dna).st_size in
  let n = lprefix + ldna in
  assert (n < Sys.max_string_length);
  let s = String.create n in
  let rec fill i cin = 
    match (try Some (input_char cin) with End_of_file -> None) with
      | Some c -> s.[i] <- c; fill (i+1) cin
      | None -> i
  in
  let m = fill (fill 0 (open_in prefix)) (open_in dna) in
  assert (m = n);
  String (0, s)

let of_string s = String (0, s)

let rec pop = function
  | String (i, s) ->
      if i >= String.length s then raise Not_found;
      base_of_char s.[i], String (i+1, s)
  | Append (_, [b], t) ->
      b, t
  | Append (n, b :: bl, t) ->
      b, Append (n-1, bl, t)
  | Append (_, [], _) ->
      assert false

let rec get t ~i = match t with
  | String (j, s) ->
      if i+j < 0 || i+j >= String.length s then raise Not_found;
      base_of_char s.[i+j]
  | Append (n, bl, t) when i < n ->
      List.nth bl i
  | Append (n, _, t) ->
      get t ~i:(i-n)

let rec chop i bl = match i, bl with
  | 0, bl -> bl
  | i, _ :: bl -> chop (i-1) bl
  | _ -> assert false

let rec skip t ~i = match t with
  | String (j, s) ->
      if i+j >= String.length s then empty else String (j+i, s)
  | Append (n, bl, t) when i < n ->
      Append (n-i, chop i bl, t)
  | Append (n, bl, t) ->
      skip t ~i:(i-n)

let baselist_of_string acc s =
  let n = String.length s in
  let rec loop acc i = 
    if i >= n then acc else loop (base_of_char s.[i] :: acc) (i+1) 
  in
  loop acc 0

let rec prefix acc i bl = match i, bl with
  | 0, _ -> acc
  | i, x :: bl -> prefix (x :: acc) (i-1) bl
  | _ -> assert false

let rec sub acc t ~start ~stop = match t with
  | String (i, s) ->
      assert (i >= 0);
      let n = String.length s in
      let index = max i (i+start) in
      let len = (min (i+stop) n) - index in
      begin 
	try baselist_of_string acc (String.sub s index len)
	with Invalid_argument _ -> acc
      end
  | Append (n, bl, t) when start >= n -> 
      (* everything in t *)
      assert (acc = []);
      sub acc t ~start:(start-n) ~stop:(stop-n)
  | Append (n, bl, t) when stop <= n -> 
      (* everything in bl *)
      let bl = chop start bl in
      prefix acc (stop - start) bl
  | Append (n, bl, t) -> 
      (* half/half *)
      let bl = chop start bl in
      let acc = prefix acc (n - start) bl in
      sub acc t ~start:0 ~stop:(stop-n)

let sub t ~start ~stop = 
  let start = max start 0 in
  if stop <= start then [] else List.rev (sub [] t ~start ~stop)

let sub t ~start ~stop = 
  let l = sub t ~start ~stop in
  Format.printf "(%d)@?" (List.length l);
  l

module T = struct
  type dna = t 
  type t = dna (* grrr *)
  let length = len
  type char = base
  let get t i = get t ~i
end
module P = struct
  type t = base array
  let length = Array.length
  type char = base
  let get = Array.get
end
module KnuthMorrisPratt = Kmp.Make(T)(P)

let search_forward t ~i ~s = 
  let t = skip t ~i in
  let p = Array.of_list s in
  let index = KnuthMorrisPratt.search t p in
  index + Array.length p

let push bl t = 
  Format.printf "[%d]@?" (List.length bl);
  if bl = [] then t else Append (List.length bl, bl, t)


let print_start length fmt dna = 
  let l = sub ~start:0 ~stop:length dna in
  List.iter (fun  b -> print_base fmt b) l 
