
{
  open Unix
  open Format
  open Lexing

  module Iset = Set.Make(struct type t = int let compare = compare end)

  let hexaval s = int_of_string ("0x" ^ s)

  let to_analyze = ref []
  let search = ref ""
  let graph = ref false
  let callers = ref ""
  let () = Arg.parse ["-search", Arg.Set_string search, "";
		      "-graph", Arg.Set graph, "";
		      "-callers", Arg.Set_string callers, ""] 
    (fun s -> to_analyze := hexaval s :: !to_analyze) ""
  let () = if !callers <> "" then graph := true

  type gene = { nbocc : int; ofs : int; len : int; seen : bool;
		mutable calls : Iset.t; 
		mutable called_by :Iset.t }

  let genes = Hashtbl.create 5033

  let add_gene nbocc ofs len seen =
    let ofs = hexaval ofs in
    Hashtbl.add genes ofs
      { nbocc = int_of_string nbocc;
	ofs = ofs;
	len = hexaval len;
	seen = seen<>""; 
	calls = Iset.empty;
	called_by = Iset.empty;
      }

  let qvalue s =
    let v = ref 0 in
    for i = 22 downto 0 do
      v := 2 * !v + (if s.[i] = 'C' then 0 else 1)
    done;
    !v

  let black = ref 0
  let red = ref 0
  let green = ref 0
  let yellow = ref 0
  let blue = ref 0
  let magenta = ref 0
  let cyan = ref 0
  let white = ref 0
  let transparent = ref 0
  let opaque = ref 0
  let reset_bucket = ref 0
  let move = ref 0
  let turn_cc = ref 0
  let turn_c = ref 0
  let mark = ref 0
  let line = ref 0
  let fill = ref 0
  let add_bitmap = ref 0
  let compose = ref 0
  let clip = ref 0
  let crossrefs = ref Iset.empty
  let add_crossref n = 
    if Hashtbl.mem genes n then
      crossrefs := Iset.add n !crossrefs

  let reset () =  
    black := 0;
    red := 0;
    green := 0;
    yellow := 0;
    blue := 0;
    magenta := 0;
    cyan := 0;
    white := 0;
    transparent := 0;
    opaque := 0;
    reset_bucket := 0;
    move := 0;
    turn_cc := 0;
    turn_c := 0;
    mark := 0;
    line := 0;
    fill := 0;
    add_bitmap := 0;
    compose := 0;
    clip := 0;
    crossrefs := Iset.empty

}

let space = [' ' '\t']*
let decimal = ['0'-'9']+
let hexa = ['0'-'9' 'a'-'f']+

let digit = 'C' | 'F'
let qinteger = digit digit digit digit digit digit digit digit digit digit digit digit digit digit digit digit digit digit digit digit digit digit digit 

rule read_genes = parse
  space (decimal as nbocc) space+ "ofs=" (hexa as ofs) space+ (hexa as len)
  space* ("//"? as seen) [^'\n']* "\n"
    { add_gene nbocc ofs len seen;
      read_genes lexbuf }
| [^'\n']* as s "\n"
    { eprintf "suspicious line: %s@." s; read_genes lexbuf }
| eof
    { () }

and scan_gene = parse
  | "IIIPIPIIIC" { incr black; scan_gene lexbuf }
  | "IIIPIPIIIP" { incr red ; scan_gene lexbuf }
  | "IIIPIPIICC" { incr green; scan_gene lexbuf }
  | "IIIPIPIICF" { incr yellow; scan_gene lexbuf }
  | "IIIPIPIICP" { incr blue; scan_gene lexbuf }
  | "IIIPIPIIFC" { incr magenta; scan_gene lexbuf }
  | "IIIPIPIIFF" { incr cyan; scan_gene lexbuf }
  | "IIIPIPIIPC" { incr white; scan_gene lexbuf }
  | "IIIPIPIIPF" { incr transparent; scan_gene lexbuf }
  | "IIIPIPIIPP" { incr opaque; scan_gene lexbuf }
  | "IIIPIIPICP" { incr reset_bucket; scan_gene lexbuf }
  | "IIIPIIIIIP" { incr move; scan_gene lexbuf }
  | "IIIPCCCCCP" { incr turn_cc; scan_gene lexbuf }
  | "IIIPFFFFFP" { incr turn_c ; scan_gene lexbuf }
  | "IIIPCCIFFP" { incr mark; scan_gene lexbuf }
  | "IIIPFFICCP" { incr line; scan_gene lexbuf }
  | "IIIPIIPIIP" { incr fill; scan_gene lexbuf }
  | "IIIPCCPFFP" { incr add_bitmap; scan_gene lexbuf }
  | "IIIPFFPCCP" { incr compose; scan_gene lexbuf }
  | "IIIPFFICCP" { incr clip; scan_gene lexbuf }
  | qinteger as v { add_crossref (qvalue v); scan_gene lexbuf }
  | _ { scan_gene lexbuf }
  | eof { () }

{
  let c = open_in "genes"
  let lb = from_channel c
  let () = read_genes lb

  let dna = 
    let n = (stat "endo.dna").st_size in
    let s = String.create n in
    let rec fill i cin = 
      match (try Some (input_char cin) with End_of_file -> None) with
	| Some c -> s.[i] <- c; fill (i+1) cin
	| None -> i
    in
    let m = fill 0 (open_in "endo.dna") in
    assert (m = n);
    s

  let dump g =
    let n = 13615 + g.ofs in
    for i = 0 to (* min 500 *) (g.len - 1) do
      printf "%c" dna.[n+i]
    done;
    printf "@."

  let () = 
    List.iter (fun gene ->
      let g = Hashtbl.find genes gene in
      eprintf "occ=%d ofs=%x len=%x seen=%b@." g.nbocc g.ofs g.len g.seen;
      dump g;
      let code = String.sub dna (13615 + g.ofs) g.len in
      reset (); scan_gene (from_string code);
      eprintf "black=%d red=%d green=%d yellow=%d@." !black !red !green !yellow;
      eprintf "blue=%d magenta=%d cyan=%d white=%d@." 
	!blue !magenta !cyan !white;
      eprintf "transparent=%d opaque=%d reset_bucket=%d@."
	!transparent !opaque !reset_bucket;
      eprintf "move=%d turn_cc=%d turn_c=%d@."
	!move !turn_cc !turn_c;
      eprintf "add_itmap=%d clip=%d compose=%d@."
	!add_bitmap !clip !compose;
      
      eprintf "%d crossrefs: " (Iset.cardinal !crossrefs);
      Iset.iter (fun c -> eprintf "%x " c) !crossrefs;
      eprintf "@.")
    !to_analyze

  let () = 
    let pat = !search in
    let npat = String.length pat in
    if pat <> "" then begin
      Hashtbl.iter
	(fun _ g -> 
	   if String.sub dna (13615 + g.ofs) npat = pat then
	     eprintf "found: ofs=%x@." g.ofs
	)
	genes
    end

  let () =
    if !graph then begin
      eprintf "building the graph...@.";
      Hashtbl.iter
	(fun _ g -> 
	   try
	     let code = String.sub dna (13615 + g.ofs) g.len in
	     reset ();
	     scan_gene (from_string code);
	     g.calls <- !crossrefs;
	     Iset.iter 
	       (fun c -> 
		  let g' = Hashtbl.find genes c in
		  g'.called_by <- Iset.add g.ofs g'.called_by) 
	       !crossrefs
	   with Invalid_argument _ ->
	     ())
	genes;
      eprintf "sorting...@.";
      let l = Hashtbl.fold (fun _ g l -> g :: l) genes [] in
      let l = 
	List.sort (fun g1 g2 -> compare (Iset.cardinal g2.called_by)
		  (Iset.cardinal g1.called_by)) l in
      let g = List.hd l in
      eprintf "most_called=%x (called %d times)@." 
	g.ofs
	(Iset.cardinal g.called_by)
    end

  let () = 
    if !callers <> "" then begin
      let g = Hashtbl.find genes (hexaval !callers) in
      eprintf "callers = ";
      Iset.iter (fun c -> eprintf "%x " c) g.called_by;
      eprintf "@."
    end

}

